@@ -34,6 +34,7 @@ extern "C" {
 #include "mrn_mysql_compat.h"
 #include <mrn_operations.hpp>
 #include <mrn_database.hpp>
+#include <mrn_buffers.hpp>
 
 #if __cplusplus >= 201402
 #  define mrn_override override
@@ -105,6 +106,17 @@ extern "C" {
 #  define MRN_NEED_FREE_STRING_MEMALLOC_PLUGIN_VAR
 #endif
 
+#if defined(MRN_MARIADB_P) || MYSQL_VERSION_ID < 80012
+#  define MRN_HAVE_HA_EXTRA_CACHE
+#  define MRN_HAVE_HA_EXTRA_NO_CACHE
+#  define MRN_HAVE_HA_EXTRA_KEY_CACHE
+#  define MRN_HAVE_HA_EXTRA_NO_KEY_CACHE
+#  define MRN_HAVE_HA_EXTRA_WRITE_CACHE
+#  define MRN_HAVE_HA_EXTRA_FLUSH_CACHE
+#  define MRN_HAVE_HA_EXTRA_REINIT_CACHE
+#  define MRN_HAVE_HA_EXTRA_MMAP
+#endif
+
 #ifdef MRN_MARIADB_P
 #  define MRN_HAVE_HA_EXTRA_DETACH_CHILD
 #  define MRN_HAVE_HA_EXTRA_PREPARE_FOR_FORCED_CLOSE
@@ -134,6 +146,12 @@ extern "C" {
 #  define MRN_HAVE_HA_EXTRA_STARTING_ORDERED_INDEX_SCAN
 #endif
 
+#if MYSQL_VERSION_ID >= 100306 && defined(MRN_MARIADB_P)
+#  define MRN_HAVE_HA_EXTRA_BEGIN_ALTER_COPY
+#  define MRN_HAVE_HA_EXTRA_END_ALTER_COPY
+#  define MRN_HAVE_HA_EXTRA_FAKE_START_STMT
+#endif
+
 #if MYSQL_VERSION_ID >= 50604 && !defined(MRN_MARIADB_P)
 #  define MRN_TIMESTAMP_USE_TIMEVAL
 #elif defined(MRN_MARIADB_P)
@@ -227,17 +245,71 @@ extern "C" {
 #  define MRN_HANDLER_HAVE_RESET_AUTO_INCREMENT
 #endif
 
+#ifdef MRN_HANDLER_HAVE_CHECK_IF_SUPPORTED_INPLACE_ALTER
+#  if defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 100306
+typedef alter_table_operations mrn_alter_flags;
+typedef alter_table_operations mrn_alter_table_flags;
+#    define MRN_ALTER_INPLACE_INFO_FLAG(alter_inplace_info_name, name)    \
+  ALTER_ ## name
+#    define MRN_ALTER_INPLACE_INFO_ALTER_FLAG(name) ALTER_ ## name
+#    define MRN_ALTER_INFO_FLAG(name) ALTER_ ## name
+#  else
+typedef Alter_inplace_info::HA_ALTER_FLAGS mrn_alter_flags;
+#    define MRN_ALTER_INPLACE_INFO_FLAG(alter_inplace_info_name, name)    \
+  alter_inplace_info_name
+#    define MRN_ALTER_INPLACE_INFO_ALTER_FLAG(name) \
+  Alter_inplace_info::ALTER_ ## name
+#    define MRN_ALTER_INFO_FLAG(name) Alter_info::ALTER_ ## name
+#  endif
+#endif
+
+#if defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 100306
+typedef alter_table_operations mrn_alter_flags;
+#else
+typedef uint mrn_alter_table_flags;
+#endif
+
 #if (!defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 50709) ||   \
   (defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 100203)
 #  define MRN_ALTER_INPLACE_INFO_ALTER_STORED_COLUMN_TYPE \
-  Alter_inplace_info::ALTER_STORED_COLUMN_TYPE
+  MRN_ALTER_INPLACE_INFO_ALTER_FLAG(STORED_COLUMN_TYPE)
 #  define MRN_ALTER_INPLACE_INFO_ALTER_STORED_COLUMN_ORDER \
-  Alter_inplace_info::ALTER_STORED_COLUMN_ORDER
+  MRN_ALTER_INPLACE_INFO_ALTER_FLAG(STORED_COLUMN_ORDER)
 #else
 #  define MRN_ALTER_INPLACE_INFO_ALTER_STORED_COLUMN_TYPE \
-  Alter_inplace_info::ALTER_COLUMN_TYPE
+  MRN_ALTER_INPLACE_INFO_ALTER_FLAG(COLUMN_TYPE)
 #  define MRN_ALTER_INPLACE_INFO_ALTER_STORED_COLUMN_ORDER \
-  Alter_inplace_info::ALTER_COLUMN_ORDER
+  MRN_ALTER_INPLACE_INFO_ALTER_FLAG(COLUMN_ORDER)
+#endif
+
+#if defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 100306
+#  define MRN_ALTER_INPLACE_INFO_ALTER_ADD_NON_UNIQUE_NON_PRIM_INDEX \
+  MRN_ALTER_INPLACE_INFO_ALTER_FLAG(ADD_NON_UNIQUE_NON_PRIM_INDEX)
+#  define MRN_ALTER_INPLACE_INFO_ALTER_DROP_NON_UNIQUE_NON_PRIM_INDEX \
+  MRN_ALTER_INPLACE_INFO_ALTER_FLAG(DROP_NON_UNIQUE_NON_PRIM_INDEX)
+#else
+#  define MRN_ALTER_INPLACE_INFO_ALTER_ADD_NON_UNIQUE_NON_PRIM_INDEX \
+  MRN_ALTER_INPLACE_INFO_FLAG(Alter_inplace_info::ADD_INDEX, ADD_INDEX)
+#  define MRN_ALTER_INPLACE_INFO_ALTER_DROP_NON_UNIQUE_NON_PRIM_INDEX \
+  MRN_ALTER_INPLACE_INFO_FLAG(Alter_inplace_info::DROP_INDEX, DROP_INDEX)
+#endif
+
+#if ((defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 100203)) || \
+  (!defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 50711)
+#  define MRN_ALTER_INPLACE_INFO_ADD_VIRTUAL_COLUMN                     \
+  MRN_ALTER_INPLACE_INFO_FLAG(Alter_inplace_info::ADD_VIRTUAL_COLUMN,   \
+                              ADD_VIRTUAL_COLUMN)
+#  define MRN_ALTER_INPLACE_INFO_ADD_STORED_BASE_COLUMN                 \
+  MRN_ALTER_INPLACE_INFO_FLAG(Alter_inplace_info::ADD_STORED_BASE_COLUMN, \
+                              ADD_STORED_BASE_COLUMN)
+#  define MRN_ALTER_INPLACE_INFO_ADD_STORED_GENERATED_COLUMN            \
+  MRN_ALTER_INPLACE_INFO_FLAG(Alter_inplace_info::ADD_STORED_GENERATED_COLUMN, \
+                              ADD_STORED_GENERATED_COLUMN)
+#else
+#  define MRN_ALTER_INPLACE_INFO_ADD_VIRTUAL_COLUMN 0
+#  define MRN_ALTER_INPLACE_INFO_ADD_STORED_BASE_COLUMN \
+  MRN_ALTER_INPLACE_INFO_FLAG(Alter_inplace_info::ADD_COLUMN, ADD_COLUMN)
+#  define MRN_ALTER_INPLACE_INFO_ADD_STORED_GENERATED_COLUMN 0
 #endif
 
 #if MYSQL_VERSION_ID >= 50700 && !defined(MRN_MARIADB_P)
@@ -248,6 +320,11 @@ extern "C" {
 #  define MRN_HANDLER_HAVE_KEYS_TO_USE_FOR_SCANNING
 #endif
 
+#if MYSQL_VERSION_ID < 80011 || defined(MRN_MARIADB_P)
+#  define MRN_HANDLER_HAVE_TABLE_CACHE_TYPE
+#  define MRN_HANDLER_HAVE_REGISTER_QUERY_CACHE_TABLE
+#endif
+
 #if (!defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 80002)
 #  define MRN_ST_MYSQL_PLUGIN_HAVE_CHECK_UNINSTALL
 #endif
@@ -257,10 +334,56 @@ extern "C" {
 #  define MRN_HANDLER_CREATE_HAVE_TABLE_DEFINITION
 #endif
 
+#if (!defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 80011)
+#  define MRN_HANDLER_DELETE_TABLE_HAVE_TABLE_DEFINITION
+#  define MRN_HANDLER_RENAME_TABLE_HAVE_TABLE_DEFINITION
+#  define MRN_HANDLER_TRUNCATE_HAVE_TABLE_DEFINITION
+#endif
+
 #if (!defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 80002)
 #  define MRN_HANDLERTON_CREATE_HAVE_PARTITIONED
 #endif
 
+#ifdef MRN_HANDLERTON_CREATE_HAVE_PARTITIONED
+#  define MRN_HANDLERTON_CREATE(hton, table, patitioned, mem_root)      \
+  (hton)->create((hton), (table), (patitioned), (mem_root))
+#else
+#  define MRN_HANDLERTON_CREATE(hton, table, patitioned, mem_root)      \
+  (hton)->create((hton), (table), (mem_root))
+#endif
+
+#if MYSQL_VERSION_ID >= 80011 && !defined(MRN_MARIADB_P)
+#  define mrn_get_new_handler(share, partitioned, alloc, db_type) \
+  get_new_handler((share), (partitioned), (alloc), (db_type))
+#else
+#  define mrn_get_new_handler(share, partitioned, alloc, db_type) \
+  get_new_handler((share), (alloc), (db_type))
+#endif
+
+#if MYSQL_VERSION_ID >= 80011 && !defined(MRN_MARIADB_P)
+#  define MRN_HANDLER_NOTIFY_TABLE_CHANGED_HAVE_ALTER_INPLACE_INFO
+#endif
+
+#if MYSQL_VERSION_ID >= 80011 && !defined(MRN_MARIADB_P)
+#  define MRN_HANDLER_PREPARE_INPLACE_ALTER_TABLE_HAVE_TABLE_DEFINITION
+#  define MRN_HANDLER_INPLACE_ALTER_TABLE_HAVE_TABLE_DEFINITION
+#  define MRN_HANDLER_COMMIT_INPLACE_ALTER_TABLE_HAVE_TABLE_DEFINITION
+#endif
+
+#if MYSQL_VERSION_ID >= 80011 && !defined(MRN_MARIADB_P)
+#  define MRN_HANDLER_HAVE_HA_MULTI_RANGE_READ_NEXT
+#  define MRN_HANDLER_HAVE_GET_REAL_ROW_TYPE
+#  define MRN_HANDLER_HAVE_GET_DEFAULT_INDEX_ALGORITHM
+#  define MRN_HANDLER_HAVE_IS_INDEX_ALGORITHM_SUPPORTED
+#  define MRN_HANDLER_HAVE_GET_SE_PRIVATE_DATA
+#  define MRN_HANDLER_HAVE_UPGRADE_TABLE
+#endif
+
+#if MYSQL_VERSION_ID >= 50631 && !defined(MRN_MARIADB_P)
+#  define MRN_HANDLER_HAVE_GET_MEMORY_BUFFER_SIZE
+#  define MRN_HANDLER_HAVE_GET_FOREIGN_DUP_KEY
+#endif
+
 #if defined(HAVE_PSI_INTERFACE) &&                      \
   (MYSQL_VERSION_ID < 80002 || defined(MRN_MARIADB_P))
 #  define MRN_HAVE_PSI_SERVER
@@ -270,6 +393,10 @@ extern "C" {
 #  define MRN_HANDLER_HA_UPDATE_ROW_NEW_DATA_CONST
 #endif
 
+#if MYSQL_VERSION_ID >= 50723 && !defined(MRN_MARIADB_P)
+#  define MRN_HANDLER_MAX_SUPPORTED_KEY_PART_LENGTH_HAVE_CREATE_INFO
+#endif
+
 #ifdef MRN_PERCONA_P
 #  define MRN_HANDLER_HAVE_HAS_GAP_LOCKS
 #endif
@@ -350,7 +477,7 @@ private:
   handler_add_index *hnd_add_index;
 #endif
 #ifdef MRN_HANDLER_HAVE_CHECK_IF_SUPPORTED_INPLACE_ALTER
-  Alter_inplace_info::HA_ALTER_FLAGS alter_handler_flags;
+  mrn_alter_flags alter_handler_flags;
   KEY         *alter_key_info_buffer;
   uint        alter_key_count;
   uint        alter_index_drop_count;
@@ -400,7 +527,7 @@ private:
   grn_table_cursor *matched_record_keys_cursor;
   grn_obj *condition_push_down_result;
   grn_table_cursor *condition_push_down_result_cursor;
-  String  *blob_buffers;
+  mrn::Buffers blob_buffers_;
 
   // for error report
   uint dup_key;
@@ -440,6 +567,9 @@ public:
              dd::Table *table_def
 #endif
     ) mrn_override;
+#ifdef MRN_HANDLER_HAVE_GET_SE_PRIVATE_DATA
+  bool get_se_private_data(dd::Table *dd_table, bool reset) mrn_override;
+#endif
   // required
   int open(const char *name, int mode, uint open_options
 #ifdef MRN_HANDLER_OPEN_HAVE_TABLE_DEFINITION
@@ -470,7 +600,12 @@ public:
   int extra(enum ha_extra_function operation);
   int extra_opt(enum ha_extra_function operation, ulong cache_size);
 
-  int delete_table(const char *name);
+  int delete_table(const char *name
+#ifdef MRN_HANDLER_DELETE_TABLE_HAVE_TABLE_DEFINITION
+                   ,
+                   const dd::Table *table_def
+#endif
+    ) mrn_override;
   int write_row(uchar *buf);
   int update_row(const uchar *old_data, const uchar *new_data);
   int update_row(const uchar *old_data, uchar *new_data);
@@ -480,7 +615,11 @@ public:
   uint max_supported_keys()            const;
   uint max_supported_key_parts()       const;
   uint max_supported_key_length()      const;
-  uint max_supported_key_part_length() const;
+  uint max_supported_key_part_length(
+#ifdef MRN_HANDLER_MAX_SUPPORTED_KEY_PART_LENGTH_HAVE_CREATE_INFO
+    HA_CREATE_INFO *create_info
+#endif
+    ) const mrn_override;
 
   ha_rows records_in_range(uint inx, key_range *min_key, key_range *max_key);
   int index_init(uint idx, bool sorted);
@@ -515,21 +654,35 @@ public:
   const Item *cond_push(const Item *cond);
   void cond_pop();
 
-  bool get_error_message(int error, String *buf);
-
   int reset();
 
-  handler *clone(const char *name, MEM_ROOT *mem_root);
+  handler *clone(const char *name, MEM_ROOT *mem_root) mrn_override;
+  void print_error(int error, myf flag) mrn_override;
+  bool get_error_message(int error, String *buffer) mrn_override;
+#ifdef MRN_HANDLER_HAVE_GET_FOREIGN_DUP_KEY
+  bool get_foreign_dup_key(char *child_table_name,
+                           uint child_table_name_len,
+                           char *child_key_name,
+                           uint child_key_name_len) mrn_override;
+#endif
+  void change_table_ptr(TABLE *table_arg, TABLE_SHARE *share_arg) mrn_override;
+  double scan_time() mrn_override;
+  double read_time(uint index, uint ranges, ha_rows rows) mrn_override;
+#ifdef MRN_HANDLER_HAVE_GET_MEMORY_BUFFER_SIZE
+  longlong get_memory_buffer_size() const mrn_override;
+#endif
+#ifdef MRN_HANDLER_HAVE_TABLE_CACHE_TYPE
   uint8 table_cache_type();
+#endif
 #ifdef MRN_HANDLER_HAVE_MULTI_RANGE_READ
   ha_rows multi_range_read_info_const(uint keyno, RANGE_SEQ_IF *seq,
                                       void *seq_init_param,
                                       uint n_ranges, uint *bufsz,
                                       uint *flags, Cost_estimate *cost);
   ha_rows multi_range_read_info(uint keyno, uint n_ranges, uint keys,
-#ifdef MRN_HANDLER_HAVE_MULTI_RANGE_READ_INFO_KEY_PARTS
+#  ifdef MRN_HANDLER_HAVE_MULTI_RANGE_READ_INFO_KEY_PARTS
                                 uint key_parts,
-#endif
+#  endif
                                 uint *bufsz, uint *flags, Cost_estimate *cost);
   int multi_range_read_init(RANGE_SEQ_IF *seq, void *seq_init_param,
                             uint n_ranges, uint mode,
@@ -549,16 +702,41 @@ public:
   void start_bulk_insert(ha_rows rows);
 #endif
   int end_bulk_insert();
+#ifdef MRN_HANDLER_HAVE_GET_SE_PRIVATE_DATA
+  bool upgrade_table(THD *thd,
+                     const char *db_name,
+                     const char *table_name,
+                     dd::Table *dd_table) mrn_override;
+#endif
   int delete_all_rows();
-  int truncate();
-  double scan_time();
-  double read_time(uint index, uint ranges, ha_rows rows);
+  int truncate(
+#ifdef MRN_HANDLER_TRUNCATE_HAVE_TABLE_DEFINITION
+    dd::Table *table_def
+#endif
+    ) mrn_override;
 #ifdef MRN_HANDLER_HAVE_KEYS_TO_USE_FOR_SCANNING
   const key_map *keys_to_use_for_scanning();
 #endif
-  ha_rows estimate_rows_upper_bound();
+  ha_rows estimate_rows_upper_bound() mrn_override;
+#ifdef MRN_HANDLER_HAVE_GET_REAL_ROW_TYPE
+  enum row_type get_real_row_type(const HA_CREATE_INFO *create_info)
+    const mrn_override;
+#endif
+#ifdef MRN_HANDLER_HAVE_GET_DEFAULT_INDEX_ALGORITHM
+  enum ha_key_alg get_default_index_algorithm() const mrn_override;
+#endif
+#ifdef MRN_HANDLER_HAVE_IS_INDEX_ALGORITHM_SUPPORTED
+  bool is_index_algorithm_supported(enum ha_key_alg key_alg) const mrn_override;
+#endif
   void update_create_info(HA_CREATE_INFO* create_info);
-  int rename_table(const char *from, const char *to);
+  int rename_table(const char *from,
+                   const char *to
+#ifdef MRN_HANDLER_RENAME_TABLE_HAVE_TABLE_DEFINITION
+                   ,
+                   const dd::Table *from_table_def,
+                   dd::Table *to_table_def
+#endif
+    ) mrn_override;
   bool is_crashed() const;
   bool auto_repair(int error) const;
   bool auto_repair() const;
@@ -577,7 +755,7 @@ public:
   check_if_supported_inplace_alter(TABLE *altered_table,
                                    Alter_inplace_info *ha_alter_info);
 #else
-  uint alter_table_flags(uint flags);
+  mrn_alter_table_flags alter_table_flags(mrn_alter_table_flags flags);
 #  ifdef MRN_HANDLER_HAVE_FINAL_ADD_INDEX
   int add_index(TABLE *table_arg, KEY *key_info, uint num_of_keys,
                 handler_add_index **add);
@@ -609,9 +787,9 @@ public:
 
 protected:
 #ifdef MRN_HANDLER_RECORDS_RETURN_ERROR
-  int records(ha_rows *num_rows);
+  int records(ha_rows *num_rows) mrn_override;
 #else
-  ha_rows records();
+  ha_rows records() mrn_override;
 #endif
 #ifdef MRN_HANDLER_HAVE_HA_RND_NEXT
   int rnd_next(uchar *buf);
@@ -636,7 +814,6 @@ protected:
 #ifdef MRN_HANDLER_HAVE_HA_INDEX_LAST
   int index_last(uchar *buf);
 #endif
-  void change_table_ptr(TABLE *table_arg, TABLE_SHARE *share_arg);
   bool primary_key_is_clustered();
   bool is_fk_defined_on_table_or_index(uint index);
   char *get_foreign_key_create_info();
@@ -653,20 +830,44 @@ protected:
   void unbind_psi();
   void rebind_psi();
 #endif
-  my_bool register_query_cache_table(THD *thd,
-                                     char *table_key,
-                                     uint key_length,
-                                     qc_engine_callback *engine_callback,
-                                     ulonglong *engine_data);
+#ifdef MRN_HANDLER_HAVE_REGISTER_QUERY_CACHE_TABLE
+  mrn_bool register_query_cache_table(THD *thd,
+                                      char *table_key,
+                                      uint key_length,
+                                      qc_engine_callback *engine_callback,
+                                      ulonglong *engine_data);
+#endif
 #ifdef MRN_HANDLER_HAVE_CHECK_IF_SUPPORTED_INPLACE_ALTER
   bool prepare_inplace_alter_table(TABLE *altered_table,
-                                   Alter_inplace_info *ha_alter_info);
+                                   Alter_inplace_info *ha_alter_info
+#  ifdef MRN_HANDLER_PREPARE_INPLACE_ALTER_TABLE_HAVE_TABLE_DEFINITION
+                                   ,
+                                   const dd::Table *old_table_def,
+                                   dd::Table *new_table_def
+#  endif
+    ) mrn_override;
   bool inplace_alter_table(TABLE *altered_table,
-                           Alter_inplace_info *ha_alter_info);
+                           Alter_inplace_info *ha_alter_info
+#  ifdef MRN_HANDLER_INPLACE_ALTER_TABLE_HAVE_TABLE_DEFINITION
+                           ,
+                           const dd::Table *old_table_def,
+                           dd::Table *new_table_def
+#  endif
+    ) mrn_override;
   bool commit_inplace_alter_table(TABLE *altered_table,
                                   Alter_inplace_info *ha_alter_info,
-                                  bool commit);
-  void notify_table_changed();
+                                  bool commit
+#  ifdef MRN_HANDLER_COMMIT_INPLACE_ALTER_TABLE_HAVE_TABLE_DEFINITION
+                                  ,
+                                  const dd::Table *old_table_def,
+                                  dd::Table *new_table_def
+#  endif
+    ) mrn_override;
+  void notify_table_changed(
+#  ifdef MRN_HANDLER_NOTIFY_TABLE_CHANGED_HAVE_ALTER_INPLACE_INFO
+    Alter_inplace_info *ha_alter_info
+#  endif
+    ) mrn_override;
 #endif
 
 private:
@@ -705,11 +906,14 @@ private:
                          grn_obj_flags *column_flags);
   grn_obj *find_column_type(Field *field, MRN_SHARE *mrn_share, int i,
                             int error_code);
-  grn_obj *find_tokenizer(KEY *key, MRN_SHARE *mrn_share, int i);
-  grn_obj *find_tokenizer(const char *name, int name_length);
+  void set_tokenizer(grn_obj *lexicon, KEY *key);
+  void set_tokenizer(grn_obj *lexicon, const char *name);
   bool have_custom_normalizer(KEY *key) const;
-  grn_obj *find_normalizer(KEY *key);
-  grn_obj *find_normalizer(KEY *key, const char *name);
+  void set_normalizer(grn_obj *lexicon, KEY *normalizer);
+  void set_normalizer(grn_obj *lexicon,
+                      KEY *key,
+                      const char *normalizer,
+                      size_t normalizer_length);
   bool find_index_column_flags(KEY *key, grn_column_flags *index_column_flags);
   bool find_token_filters(KEY *key, grn_obj *token_filters);
   bool find_token_filters_put(grn_obj *token_filters,
@@ -729,7 +933,7 @@ private:
 #endif
   bool is_dry_write();
   bool is_enable_optimization();
-  bool should_normalize(Field *field) const;
+  bool should_normalize(Field *field, bool is_fulltext_index) const;
   void check_count_skip(key_part_map target_key_part_map);
   bool is_grn_zero_column_value(grn_obj *column, grn_obj *value);
   bool is_primary_key_field(Field *field) const;
@@ -848,7 +1052,11 @@ private:
                               uchar *buf, uint *size);
   int storage_encode_key_set(Field *field, const uchar *key,
                              uchar *buf, uint *size);
-  int storage_encode_key(Field *field, const uchar *key, uchar *buf, uint *size);
+  int storage_encode_key(Field *field,
+                         const uchar *key,
+                         uchar *buf,
+                         uint *size,
+                         mrn_bool *is_null);
   int storage_encode_multiple_column_key(KEY *key_info,
                                          const uchar *key, uint key_length,
                                          uchar *buffer, uint *encoded_length);
@@ -871,39 +1079,57 @@ private:
 
   void set_pk_bitmap();
   int create_share_for_create() const;
-  int wrapper_create(const char *name, TABLE *table,
-                     HA_CREATE_INFO *info, MRN_SHARE *tmp_share);
-  int storage_create(const char *name, TABLE *table,
-                     HA_CREATE_INFO *info, MRN_SHARE *tmp_share);
+  int wrapper_create(const char *name,
+                     TABLE *table,
+                     HA_CREATE_INFO *info,
+#ifdef MRN_HANDLER_OPEN_HAVE_TABLE_DEFINITION
+                     dd::Table *table_def,
+#endif
+                     MRN_SHARE *tmp_share);
+  int storage_create(const char *name,
+                     TABLE *table,
+                     HA_CREATE_INFO *info,
+#ifdef MRN_HANDLER_OPEN_HAVE_TABLE_DEFINITION
+                     dd::Table *table_def,
+#endif
+                     MRN_SHARE *tmp_share);
+#ifdef MRN_HANDLER_HAVE_GET_SE_PRIVATE_DATA
+  bool wrapper_get_se_private_data(dd::Table *dd_table, bool reset);
+  bool storage_get_se_private_data(dd::Table *dd_table, bool reset);
+#endif
   int wrapper_create_index_fulltext_validate(KEY *key_info);
   int wrapper_create_index_fulltext(const char *grn_table_name,
                                     int i,
                                     KEY *key_info,
                                     grn_obj **index_tables,
-                                    grn_obj **index_columns,
-                                    MRN_SHARE *tmp_share);
+                                    grn_obj **index_columns);
   int wrapper_create_index_geo(const char *grn_table_name,
                                int i,
                                KEY *key_info,
                                grn_obj **index_tables,
-                               grn_obj **index_columns,
-                               MRN_SHARE *tmp_share);
+                               grn_obj **index_columns);
   int wrapper_create_index(const char *name,
                            TABLE *table,
                            HA_CREATE_INFO *info,
                            MRN_SHARE *tmp_share);
   int storage_create_validate_pseudo_column(TABLE *table);
 #ifdef MRN_SUPPORT_FOREIGN_KEYS
-  bool storage_create_foreign_key(TABLE *table, const char *grn_table_name,
-                                  Field *field, grn_obj *table_obj, int &error);
+  bool storage_create_foreign_key(TABLE *table,
+                                  const char *grn_table_name,
+                                  Field *field,
+                                  grn_obj *table_obj,
+#  ifdef MRN_OPEN_TABLE_DEF_USE_TABLE_DEFINITION
+                                  const dd::Table *table_def,
+#  endif
+                                  int &error);
 #endif
   int storage_create_validate_index(TABLE *table);
   int storage_create_index_table(TABLE *table, const char *grn_table_name,
-                                 grn_obj *grn_table, MRN_SHARE *tmp_share,
+                                 grn_obj *grn_table,
                                  KEY *key_info, grn_obj **index_tables,
                                  uint i);
   int storage_create_index(TABLE *table, const char *grn_table_name,
-                           grn_obj *grn_table, MRN_SHARE *tmp_share,
+                           grn_obj *grn_table,
                            KEY *key_info, grn_obj **index_tables,
                            grn_obj **index_columns, uint i);
   int storage_create_indexes(TABLE *table, const char *grn_table_name,
@@ -911,13 +1137,35 @@ private:
   int close_databases();
   int ensure_database_open(const char *name, mrn::Database **db=NULL);
   int ensure_database_remove(const char *name);
-  int wrapper_delete_table(const char *name, handlerton *wrap_handlerton,
+  int wrapper_delete_table(const char *name,
+#ifdef MRN_HANDLER_DELETE_TABLE_HAVE_TABLE_DEFINITION
+                           const dd::Table *table_def,
+#endif
+                           handlerton *wrap_handlerton,
                            const char *table_name);
-  int generic_delete_table(const char *name, const char *table_name);
-  int wrapper_open(const char *name, int mode, uint open_options);
+  int generic_delete_table(const char *name,
+#ifdef MRN_HANDLER_DELETE_TABLE_HAVE_TABLE_DEFINITION
+                           const dd::Table *table_def,
+#endif
+                           const char *table_name);
+  int wrapper_open(const char *name,
+                   int mode,
+                   uint open_options
+#ifdef MRN_HANDLER_OPEN_HAVE_TABLE_DEFINITION
+                   ,
+                   const dd::Table *table_def
+#endif
+    );
   int wrapper_open_indexes(const char *name);
   int storage_reindex();
-  int storage_open(const char *name, int mode, uint open_options);
+  int storage_open(const char *name,
+                   int mode,
+                   uint open_options
+#ifdef MRN_HANDLER_OPEN_HAVE_TABLE_DEFINITION
+                   ,
+                   const dd::Table *table_def
+#endif
+    );
   int open_table(const char *name);
   int storage_open_columns(void);
   void storage_close_columns(void);
@@ -950,6 +1198,16 @@ private:
 #endif
   int wrapper_end_bulk_insert();
   int storage_end_bulk_insert();
+#ifdef MRN_HANDLER_HAVE_GET_SE_PRIVATE_DATA
+  bool wrapper_upgrade_table(THD *thd,
+                             const char *db_name,
+                             const char *table_name,
+                             dd::Table *dd_table);
+  bool storage_upgrade_table(THD *thd,
+                             const char *db_name,
+                             const char *table_name,
+                             dd::Table *dd_table);
+#endif
   bool wrapper_is_target_index(KEY *key_info);
   bool wrapper_have_target_index();
   int wrapper_write_row(uchar *buf);
@@ -994,8 +1252,16 @@ private:
   uint storage_max_supported_key_parts() const;
   uint wrapper_max_supported_key_length() const;
   uint storage_max_supported_key_length() const;
-  uint wrapper_max_supported_key_part_length() const;
-  uint storage_max_supported_key_part_length() const;
+  uint wrapper_max_supported_key_part_length(
+#ifdef MRN_HANDLER_MAX_SUPPORTED_KEY_PART_LENGTH_HAVE_CREATE_INFO
+    HA_CREATE_INFO *create_info
+#endif
+    ) const;
+  uint storage_max_supported_key_part_length(
+#ifdef MRN_HANDLER_MAX_SUPPORTED_KEY_PART_LENGTH_HAVE_CREATE_INFO
+    HA_CREATE_INFO *create_info
+#endif
+    ) const;
   ulonglong wrapper_table_flags() const;
   ulonglong storage_table_flags() const;
   ulong wrapper_index_flags(uint idx, uint part, bool all_parts) const;
@@ -1083,12 +1349,36 @@ private:
   const Item *storage_cond_push(const Item *cond);
   void wrapper_cond_pop();
   void storage_cond_pop();
-  bool wrapper_get_error_message(int error, String *buf);
-  bool storage_get_error_message(int error, String *buf);
   handler *wrapper_clone(const char *name, MEM_ROOT *mem_root);
   handler *storage_clone(const char *name, MEM_ROOT *mem_root);
+  void wrapper_print_error(int error, myf flag);
+  void storage_print_error(int error, myf flag);
+  bool wrapper_get_error_message(int error, String *buffer);
+  bool storage_get_error_message(int error, String *buffer);
+#ifdef MRN_HANDLER_HAVE_GET_FOREIGN_DUP_KEY
+  bool wrapper_get_foreign_dup_key(char *child_table_name,
+                                   uint child_table_name_len,
+                                   char *child_key_name,
+                                   uint child_key_name_len);
+  bool storage_get_foreign_dup_key(char *child_table_name,
+                                   uint child_table_name_len,
+                                   char *child_key_name,
+                                   uint child_key_name_len);
+#endif
+  void wrapper_change_table_ptr(TABLE *table_arg, TABLE_SHARE *share_arg);
+  void storage_change_table_ptr(TABLE *table_arg, TABLE_SHARE *share_arg);
+  double wrapper_scan_time();
+  double storage_scan_time();
+  double wrapper_read_time(uint index, uint ranges, ha_rows rows);
+  double storage_read_time(uint index, uint ranges, ha_rows rows);
+#ifdef MRN_HANDLER_HAVE_GET_MEMORY_BUFFER_SIZE
+  longlong wrapper_get_memory_buffer_size() const;
+  longlong storage_get_memory_buffer_size() const;
+#endif
+#ifdef MRN_HANDLER_HAVE_TABLE_CACHE_TYPE
   uint8 wrapper_table_cache_type();
   uint8 storage_table_cache_type();
+#endif
 #ifdef MRN_HANDLER_HAVE_MULTI_RANGE_READ
   ha_rows wrapper_multi_range_read_info_const(uint keyno,
                                               RANGE_SEQ_IF *seq,
@@ -1142,26 +1432,50 @@ private:
                               const char *function_name);
   int wrapper_delete_all_rows();
   int storage_delete_all_rows();
-  int wrapper_truncate();
+  int wrapper_truncate(
+#ifdef MRN_HANDLER_TRUNCATE_HAVE_TABLE_DEFINITION
+    dd::Table *table_def
+#endif
+    );
   int wrapper_truncate_index();
-  int storage_truncate();
+  int storage_truncate(
+#ifdef MRN_HANDLER_TRUNCATE_HAVE_TABLE_DEFINITION
+    dd::Table *table_def
+#endif
+    );
   int storage_truncate_index();
-  double wrapper_scan_time();
-  double storage_scan_time();
-  double wrapper_read_time(uint index, uint ranges, ha_rows rows);
-  double storage_read_time(uint index, uint ranges, ha_rows rows);
 #ifdef MRN_HANDLER_HAVE_KEYS_TO_USE_FOR_SCANNING
   const key_map *wrapper_keys_to_use_for_scanning();
   const key_map *storage_keys_to_use_for_scanning();
 #endif
   ha_rows wrapper_estimate_rows_upper_bound();
   ha_rows storage_estimate_rows_upper_bound();
+#ifdef MRN_HANDLER_HAVE_GET_REAL_ROW_TYPE
+  enum row_type wrapper_get_real_row_type(const HA_CREATE_INFO *create_info)
+    const;
+  enum row_type storage_get_real_row_type(const HA_CREATE_INFO *create_info)
+    const;
+#endif
+#ifdef MRN_HANDLER_HAVE_GET_DEFAULT_INDEX_ALGORITHM
+  enum ha_key_alg wrapper_get_default_index_algorithm() const;
+  enum ha_key_alg storage_get_default_index_algorithm() const;
+#endif
+#ifdef MRN_HANDLER_HAVE_IS_INDEX_ALGORITHM_SUPPORTED
+  bool wrapper_is_index_algorithm_supported(enum ha_key_alg key_alg) const;
+  bool storage_is_index_algorithm_supported(enum ha_key_alg key_alg) const;
+#endif
   void wrapper_update_create_info(HA_CREATE_INFO* create_info);
   void storage_update_create_info(HA_CREATE_INFO* create_info);
   int wrapper_rename_table(const char *from, const char *to,
                            MRN_SHARE *tmp_share,
                            const char *from_table_name,
-                           const char *to_table_name);
+                           const char *to_table_name
+#ifdef MRN_HANDLER_RENAME_TABLE_HAVE_TABLE_DEFINITION
+                           ,
+                           const dd::Table *from_table_def,
+                           dd::Table *to_table_def
+#endif
+    );
   int wrapper_rename_index(const char *from, const char *to,
                            MRN_SHARE *tmp_share,
                            const char *from_table_name,
@@ -1169,7 +1483,13 @@ private:
   int storage_rename_table(const char *from, const char *to,
                            MRN_SHARE *tmp_share,
                            const char *from_table_name,
-                           const char *to_table_name);
+                           const char *to_table_name
+#ifdef MRN_HANDLER_RENAME_TABLE_HAVE_TABLE_DEFINITION
+                           ,
+                           const dd::Table *from_table_def,
+                           dd::Table *to_table_def
+#endif
+    );
 #ifdef MRN_SUPPORT_FOREIGN_KEYS
   int storage_rename_foreign_key(MRN_SHARE *tmp_share,
                                  const char *from_table_name,
@@ -1219,11 +1539,29 @@ private:
   storage_check_if_supported_inplace_alter(TABLE *altered_table,
                                            Alter_inplace_info *ha_alter_info);
   bool wrapper_prepare_inplace_alter_table(TABLE *altered_table,
-                                           Alter_inplace_info *ha_alter_info);
+                                           Alter_inplace_info *ha_alter_info
+#  ifdef MRN_HANDLER_PREPARE_INPLACE_ALTER_TABLE_HAVE_TABLE_DEFINITION
+                                           ,
+                                           const dd::Table *old_table_def,
+                                           dd::Table *new_table_def
+#  endif
+    );
   bool storage_prepare_inplace_alter_table(TABLE *altered_table,
-                                           Alter_inplace_info *ha_alter_info);
+                                           Alter_inplace_info *ha_alter_info
+#  ifdef MRN_HANDLER_PREPARE_INPLACE_ALTER_TABLE_HAVE_TABLE_DEFINITION
+                                           ,
+                                           const dd::Table *old_table_def,
+                                           dd::Table *new_table_def
+#  endif
+    );
   bool wrapper_inplace_alter_table(TABLE *altered_table,
-                                   Alter_inplace_info *ha_alter_info);
+                                   Alter_inplace_info *ha_alter_info
+#  ifdef MRN_HANDLER_COMMIT_INPLACE_ALTER_TABLE_HAVE_TABLE_DEFINITION
+                                   ,
+                                   const dd::Table *old_table_def,
+                                   dd::Table *new_table_def
+#  endif
+    );
   bool storage_inplace_alter_table_add_index(TABLE *altered_table,
                                              Alter_inplace_info *ha_alter_info);
   bool storage_inplace_alter_table_drop_index(TABLE *altered_table,
@@ -1235,18 +1573,44 @@ private:
   bool storage_inplace_alter_table_rename_column(TABLE *altered_table,
                                                  Alter_inplace_info *ha_alter_info);
   bool storage_inplace_alter_table(TABLE *altered_table,
-                                   Alter_inplace_info *ha_alter_info);
+                                   Alter_inplace_info *ha_alter_info
+#  ifdef MRN_HANDLER_COMMIT_INPLACE_ALTER_TABLE_HAVE_TABLE_DEFINITION
+                                   ,
+                                   const dd::Table *old_table_def,
+                                   dd::Table *new_table_def
+#  endif
+    );
   bool wrapper_commit_inplace_alter_table(TABLE *altered_table,
                                           Alter_inplace_info *ha_alter_info,
-                                          bool commit);
+                                          bool commit
+#  ifdef MRN_HANDLER_COMMIT_INPLACE_ALTER_TABLE_HAVE_TABLE_DEFINITION
+                                          ,
+                                          const dd::Table *old_table_def,
+                                          dd::Table *new_table_def
+#  endif
+    );
   bool storage_commit_inplace_alter_table(TABLE *altered_table,
                                           Alter_inplace_info *ha_alter_info,
-                                          bool commit);
-  void wrapper_notify_table_changed();
-  void storage_notify_table_changed();
+                                          bool commit
+#  ifdef MRN_HANDLER_COMMIT_INPLACE_ALTER_TABLE_HAVE_TABLE_DEFINITION
+                                          ,
+                                          const dd::Table *old_table_def,
+                                          dd::Table *new_table_def
+#  endif
+    );
+  void wrapper_notify_table_changed(
+#  ifdef MRN_HANDLER_NOTIFY_TABLE_CHANGED_HAVE_ALTER_INPLACE_INFO
+    Alter_inplace_info *ha_alter_info
+#  endif
+    );
+  void storage_notify_table_changed(
+#  ifdef MRN_HANDLER_NOTIFY_TABLE_CHANGED_HAVE_ALTER_INPLACE_INFO
+    Alter_inplace_info *ha_alter_info
+#  endif
+    );
 #else
-  uint wrapper_alter_table_flags(uint flags);
-  uint storage_alter_table_flags(uint flags);
+  mrn_alter_table_flags wrapper_alter_table_flags(mrn_alter_table_flags flags);
+  mrn_alter_table_flags storage_alter_table_flags(mrn_alter_table_flags flags);
 #  ifdef MRN_HANDLER_HAVE_FINAL_ADD_INDEX
   int wrapper_add_index(TABLE *table_arg, KEY *key_info, uint num_of_keys,
                         handler_add_index **add);
@@ -1301,8 +1665,6 @@ private:
   bool wrapper_has_gap_locks() const;
   bool storage_has_gap_locks() const;
 #endif
-  void wrapper_change_table_ptr(TABLE *table_arg, TABLE_SHARE *share_arg);
-  void storage_change_table_ptr(TABLE *table_arg, TABLE_SHARE *share_arg);
   bool wrapper_primary_key_is_clustered();
   bool storage_primary_key_is_clustered();
   bool wrapper_is_fk_defined_on_table_or_index(uint index);
@@ -1336,18 +1698,20 @@ private:
   void wrapper_rebind_psi();
   void storage_rebind_psi();
 #endif
-  my_bool wrapper_register_query_cache_table(THD *thd,
-                                             char *table_key,
-                                             uint key_length,
-                                             qc_engine_callback
-                                             *engine_callback,
-                                             ulonglong *engine_data);
-  my_bool storage_register_query_cache_table(THD *thd,
-                                             char *table_key,
-                                             uint key_length,
-                                             qc_engine_callback
-                                             *engine_callback,
-                                             ulonglong *engine_data);
+#ifdef MRN_HANDLER_HAVE_REGISTER_QUERY_CACHE_TABLE
+  mrn_bool wrapper_register_query_cache_table(THD *thd,
+                                              char *table_key,
+                                              uint key_length,
+                                              qc_engine_callback
+                                              *engine_callback,
+                                              ulonglong *engine_data);
+  mrn_bool storage_register_query_cache_table(THD *thd,
+                                              char *table_key,
+                                              uint key_length,
+                                              qc_engine_callback
+                                              *engine_callback,
+                                              ulonglong *engine_data);
+#endif
 };
 
 #ifdef __cplusplus
