@@ -1,7 +1,7 @@
 /* -*- c-basic-offset: 2 -*- */
 /*
   Copyright(C) 2011 Kentoku SHIBA
-  Copyright(C) 2011-2012 Kouhei Sutou <kou@clear-code.com>
+  Copyright(C) 2011-2017 Kouhei Sutou <kou@clear-code.com>
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -15,7 +15,7 @@
 
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
-  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include <mrn_mysql.h>
@@ -26,19 +26,72 @@
 #define MRN_CLASS_NAME "mrn::IndexTableName"
 
 namespace mrn {
+  const char *IndexTableName::SEPARATOR = "#";
+  const char *IndexTableName::OLD_SEPARATOR = "-";
+
+  bool IndexTableName::is_custom_name(const char *table_name,
+                                      size_t table_name_length,
+                                      const char *index_table_name,
+                                      size_t index_table_name_length)
+  {
+    MRN_DBUG_ENTER_METHOD();
+
+    if (index_table_name_length <= (table_name_length + strlen(SEPARATOR))) {
+      DBUG_RETURN(true);
+    }
+
+    if (strncmp(table_name, index_table_name, table_name_length) != 0) {
+      DBUG_RETURN(true);
+    }
+
+    if ((strncmp(OLD_SEPARATOR,
+                 index_table_name + table_name_length,
+                 strlen(OLD_SEPARATOR)) != 0) &&
+        (strncmp(SEPARATOR,
+                 index_table_name + table_name_length,
+                 strlen(SEPARATOR)) != 0)) {
+      DBUG_RETURN(true);
+    }
+
+    DBUG_RETURN(false);
+  }
+
   IndexTableName::IndexTableName(const char *table_name,
                                  const char *mysql_index_name)
     : table_name_(table_name),
-      mysql_index_name_(mysql_index_name) {
+      mysql_index_name_(mysql_index_name),
+      mysql_index_name_length_(strlen(mysql_index_name_)) {
+    init();
+  }
+
+  IndexTableName::IndexTableName(const char *table_name,
+                                 const char *mysql_index_name,
+                                 size_t mysql_index_name_length)
+    : table_name_(table_name),
+      mysql_index_name_(mysql_index_name),
+      mysql_index_name_length_(mysql_index_name_length) {
+    init();
+  }
+
+  void IndexTableName::init() {
     uchar encoded_mysql_index_name_multibyte[MRN_MAX_KEY_SIZE];
     const uchar *mysql_index_name_multibyte =
       reinterpret_cast<const uchar *>(mysql_index_name_);
     encode(encoded_mysql_index_name_multibyte,
            encoded_mysql_index_name_multibyte + MRN_MAX_KEY_SIZE,
            mysql_index_name_multibyte,
-           mysql_index_name_multibyte + strlen(mysql_index_name_));
+           mysql_index_name_multibyte + mysql_index_name_length_);
+    snprintf(old_name_, MRN_MAX_KEY_SIZE,
+             "%s%s%s",
+             table_name_,
+             OLD_SEPARATOR,
+             encoded_mysql_index_name_multibyte);
+    old_length_ = strlen(old_name_);
     snprintf(name_, MRN_MAX_KEY_SIZE,
-             "%s-%s", table_name_, encoded_mysql_index_name_multibyte);
+             "%s%s%s",
+             table_name_,
+             SEPARATOR,
+             encoded_mysql_index_name_multibyte);
     length_ = strlen(name_);
   }
 
@@ -50,6 +103,14 @@ namespace mrn {
     return length_;
   }
 
+  const char *IndexTableName::old_c_str() {
+    return old_name_;
+  }
+
+  size_t IndexTableName::old_length() {
+    return old_length_;
+  }
+
   uint IndexTableName::encode(uchar *encoded_start,
                               uchar *encoded_end,
                               const uchar *mysql_string_start,
