@@ -1,7 +1,7 @@
 /* -*- c-basic-offset: 2 -*- */
 /*
   Copyright(C) 2011-2013 Kentoku SHIBA
-  Copyright(C) 2011-2013 Kouhei Sutou <kou@clear-code.com>
+  Copyright(C) 2011-2018 Kouhei Sutou <kou@clear-code.com>
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -15,12 +15,17 @@
 
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
-  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include "mrn_parameters_parser.hpp"
 
 #include <mrn_mysql_compat.h>
+#include <mrn_variables.hpp>
+
+#ifdef MRN_HAVE_SQL_DERROR_H
+#  include <sql/derror.h>
+#endif
 
 namespace mrn {
   class Parameter {
@@ -30,15 +35,15 @@ namespace mrn {
 
     Parameter(const char *key, unsigned int key_length,
               const char *value, unsigned int value_length)
-      : key_(my_strndup(key, key_length, MYF(0))),
-        value_(my_strndup(value, value_length, MYF(0))) {
+      : key_(mrn_my_strndup(key, key_length, MYF(0))),
+        value_(mrn_my_strndup(value, value_length, MYF(0))) {
     };
     ~Parameter() {
       if (key_) {
-        my_free(key_, MYF(0));
+        my_free(key_);
       }
       if (value_) {
-        my_free(value_, MYF(0));
+        my_free(value_);
       }
     };
   };
@@ -173,4 +178,24 @@ namespace mrn {
     }
     return NULL;
   }
+
+  const char *ParametersParser::tokenizer() {
+    const char *parser = (*this)["parser"];
+    if (parser) {
+      push_warning_printf(current_thd,
+                          MRN_SEVERITY_WARNING,
+                          ER_WARN_DEPRECATED_SYNTAX,
+                          MRN_GET_ERR_MSG(ER_WARN_DEPRECATED_SYNTAX),
+                          "parser", "tokenizer");
+    }
+    const char *tokenizer = (*this)["tokenizer"];
+    if (!tokenizer) {
+      tokenizer = parser;
+    }
+    return tokenizer;
+  }
+
+  const char *ParametersParser::table() {
+    return (*this)["table"];
+  }
 }
