@@ -83,7 +83,8 @@
 #endif
 
 #ifdef MRN_KEY_NAME_IS_LEX_STRING
-#  define KEY_NAME(key) (key)->name.str, (key)->name.length
+#  define KEY_NAME_PTR(key) ((key)->name.str)
+#  define KEY_NAME_LENGTH(key) ((key)->name.length)
 #  define KEY_NAME_EQUAL_KEY(key1, key2)                \
   ((key1)->name.length == (key2)->name.length &&        \
    strncmp((key1)->name.str,                            \
@@ -94,12 +95,14 @@
   static_cast<int>((key)->name.length),         \
   (key)->name.str
 #else
-#  define KEY_NAME(key) (key)->name, strlen((key)->name)
+#  define KEY_NAME_PTR(key) (key)->name
+#  define KEY_NAME_LENGTH(key) strlen((key)->name)
 #  define KEY_NAME_EQUAL_KEY(key1, key2)        \
   strcmp((key1)->name, (key2)->name) == 0
 #  define KEY_NAME_FORMAT "%s"
 #  define KEY_NAME_FORMAT_VALUE(key) (key)->name
 #endif
+#define KEY_NAME(key) KEY_NAME_PTR(key), KEY_NAME_LENGTH(key)
 
 #ifdef MRN_FIELD_FIELD_NAME_IS_LEX_STRING
 #  define FIELD_NAME(field)                             \
@@ -120,7 +123,7 @@
                       NULL,                                             \
                       (field)->field_name.str,                          \
                       (field)->field_name.length,                       \
-                      TRUE);
+                      true);
 #  define FIELD_NAME_FORMAT "%.*s"
 #  define FIELD_NAME_FORMAT_VALUE(field)                                \
   static_cast<int>((field)->field_name.length), (field)->field_name.str
@@ -139,12 +142,13 @@
                        NULL,                            \
                        (field)->field_name,             \
                        strlen((field)->field_name),     \
-                       TRUE);
+                       true);
 #  define FIELD_NAME_FORMAT "%s"
 #  define FIELD_NAME_FORMAT_VALUE(field) (field)->field_name
 #endif
 
-#if MYSQL_VERSION_ID >= 100302 && defined(MRN_MARIADB_P)
+#if (MYSQL_VERSION_ID >= 100302 && defined(MRN_MARIADB_P)) ||   \
+  (MYSQL_VERSION_ID >= 80011 && !defined(MRN_MARIADB_P))
 #  define MRN_KEY_PART_SPEC_FIELD_NAME_USE_CONST_STRING
 #endif
 
@@ -164,15 +168,18 @@
   typedef MYSQL_LEX_STRING mrn_thd_lex_string;
 #endif
 
-#if defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 100000
-#  define mrn_init_alloc_root(PTR, SZ1, SZ2, FLAG) \
-  init_alloc_root(PTR, SZ1, SZ2, FLAG)
+#if defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 100306
+#  define mrn_init_alloc_root(root, name, block_size, pre_alloc_size, flags) \
+  init_alloc_root(root, name, block_size, pre_alloc_size, flags)
+#elif defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 100000
+#  define mrn_init_alloc_root(root, name, block_size, pre_alloc_size, flags) \
+  init_alloc_root(root, block_size, pre_alloc_size, flags)
 #elif MYSQL_VERSION_ID >= 50706
-#  define mrn_init_alloc_root(PTR, SZ1, SZ2, FLAG) \
-  init_alloc_root(mrn_memory_key, PTR, SZ1, SZ2)
+#  define mrn_init_alloc_root(root, name, block_size, pre_alloc_size, flags) \
+  init_alloc_root(mrn_memory_key, root, block_size, pre_alloc_size)
 #else
-#  define mrn_init_alloc_root(PTR, SZ1, SZ2, FLAG) \
-  init_alloc_root(PTR, SZ1, SZ2)
+#  define mrn_init_alloc_root(root, name, block_size, pre_alloc_size, flags) \
+  init_alloc_root(root, block_size, pre_alloc_size)
 #endif
 
 #if MYSQL_VERSION_ID < 100002 || !defined(MRN_MARIADB_P)
@@ -211,6 +218,19 @@
 
 #if MYSQL_VERSION_ID < 100000
 #  define MRN_HAVE_TABLE_DEF_CACHE
+#  if MYSQL_VERSION_ID >= 80011 && !defined(MRN_MARIADB_P)
+#    define MRN_TABLE_DEF_CACHE_TYPE_IS_MAP
+#  endif
+
+#  ifdef MRN_TABLE_DEF_CACHE_TYPE_IS_MAP
+struct Table_share_deleter;
+typedef
+  malloc_unordered_map<std::string,
+                       std::unique_ptr<TABLE_SHARE, Table_share_deleter>> *
+  mrn_table_def_cache_type;
+#  else
+typedef HASH mrn_table_def_cache_type;
+#  endif
 #endif
 
 #if defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 100009
@@ -315,6 +335,14 @@
 #  define MRN_HAVE_BINLOG_H
 #endif
 
+#if MYSQL_VERSION_ID >= 50720 && !defined(MRN_MARIADB_P)
+#  define MRN_HAVE_MYSQL_PSI_MYSQL_MEMORY_H
+#endif
+
+#if MYSQL_VERSION_ID >= 80011 && !defined(MRN_MARIADB_P)
+#  define MRN_HAVE_SQL_DERROR_H
+#endif
+
 #if MYSQL_VERSION_ID >= 50706 && !defined(MRN_MARIADB_P)
 #  define MRN_HAVE_SPATIAL
 #elif defined(HAVE_SPATIAL)
@@ -400,43 +428,6 @@
 #define MRN_ERROR_CODE_DATA_TRUNCATE(thd)                               \
   (MRN_ABORT_ON_WARNING(thd) ? ER_WARN_DATA_OUT_OF_RANGE : WARN_DATA_TRUNCATED)
 
-#if MYSQL_VERSION_ID >= 50709 && !defined(MRN_MARIADB_P)
-#  define mrn_my_hash_init(hash,                        \
-                           charset,                     \
-                           default_array_elements,      \
-                           key_offset,                  \
-                           key_length,                  \
-                           get_key,                     \
-                           free_element,                \
-                           flags)                       \
-  my_hash_init(hash,                                    \
-               charset,                                 \
-               default_array_elements,                  \
-               key_offset,                              \
-               key_length,                              \
-               get_key,                                 \
-               free_element,                            \
-               flags,                                   \
-               mrn_memory_key)
-#else
-#  define mrn_my_hash_init(hash,                        \
-                           charset,                     \
-                           default_array_elements,      \
-                           key_offset,                  \
-                           key_length,                  \
-                           get_key,                     \
-                           free_element,                \
-                           flags)                       \
-  my_hash_init(hash,                                    \
-               charset,                                 \
-               default_array_elements,                  \
-               key_offset,                              \
-               key_length,                              \
-               get_key,                                 \
-               free_element,                            \
-               flags)
-#endif
-
 #if defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 100000
 #  define mrn_strconvert(from_cs,               \
                          from,                  \
@@ -539,21 +530,6 @@
 #define MRN_FREE_VARIABLE_LENGTH_ARRAYS(variable_name) \
   my_free(variable_name)
 
-#if ((defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 100203)) || \
-  (!defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 50711)
-#  define MRN_ALTER_INPLACE_INFO_ADD_VIRTUAL_COLUMN \
-  Alter_inplace_info::ADD_VIRTUAL_COLUMN
-#  define MRN_ALTER_INPLACE_INFO_ADD_STORED_BASE_COLUMN \
-  Alter_inplace_info::ADD_STORED_BASE_COLUMN
-#  define MRN_ALTER_INPLACE_INFO_ADD_STORED_GENERATED_COLUMN \
-  Alter_inplace_info::ADD_STORED_GENERATED_COLUMN
-#else
-#  define MRN_ALTER_INPLACE_INFO_ADD_VIRTUAL_COLUMN 0
-#  define MRN_ALTER_INPLACE_INFO_ADD_STORED_BASE_COLUMN \
-  Alter_inplace_info::ADD_COLUMN
-#  define MRN_ALTER_INPLACE_INFO_ADD_STORED_GENERATED_COLUMN 0
-#endif
-
 #if (defined(HA_CAN_VIRTUAL_COLUMNS) || defined(HA_GENERATED_COLUMNS))
 #  define MRN_SUPPORT_GENERATED_COLUMNS
 #endif
@@ -606,3 +582,130 @@
 #    define MRN_HAVE_DB_TYPE_ROCKSDB
 #  endif
 #endif
+
+#if defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 100306
+#  define MRN_TABLE_LIST_TABLE_NAME_DATA(table_list)    \
+  (table_list)->table_name.str
+#  define MRN_TABLE_LIST_TABLE_NAME_LENGTH(table_list)  \
+  (table_list)->table_name.length
+#else
+#  define MRN_TABLE_LIST_TABLE_NAME_DATA(table_list)    \
+  (table_list)->table_name
+#  define MRN_TABLE_LIST_TABLE_NAME_LENGTH(table_list)  \
+  (table_list)->table_name_length
+#endif
+
+#if defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 100306
+#  define mrn_table_list_init_one_table(table_list,                     \
+                                        db_name,                        \
+                                        db_name_length,                 \
+                                        table_name,                     \
+                                        table_name_length,              \
+                                        alias,                          \
+                                        lock_type) do {                 \
+    LEX_CSTRING db_name_ = {db_name, db_name_length};                   \
+    LEX_CSTRING table_name_ = {table_name, table_name_length};          \
+    LEX_CSTRING alias_ = {alias, strlen(alias)};                        \
+    (table_list)->init_one_table(&db_name_,                             \
+                                 &table_name_,                          \
+                                 &alias_,                               \
+                                 lock_type);                            \
+  } while(false)
+#else
+#  define mrn_table_list_init_one_table(table_list,             \
+                                        db_name,                \
+                                        db_name_length,         \
+                                        table_name,             \
+                                        table_name_length,      \
+                                        alias,                  \
+                                        lock_type)              \
+  (table_list)->init_one_table(db_name,                         \
+                               db_name_length,                  \
+                               table_name,                      \
+                               table_name_length,               \
+                               alias,                           \
+                               lock_type)
+#endif
+
+#if MYSQL_VERSION_ID >= 80011 && !defined(MRN_MARIADB_P)
+  typedef Key_spec mrn_key_spec;
+#  define MRN_KEY_SPEC_LIST_EACH_BEGIN(spec_list, spec) do {            \
+    for (size_t spec_i = 0; spec_i < spec_list.size(); ++spec_i) {      \
+      const mrn_key_spec *spec = spec_list[spec_i];
+#  define MRN_KEY_SPEC_LIST_EACH_END()          \
+    }                                           \
+  } while (false)
+#else
+  typedef Key mrn_key_spec;
+#  define MRN_KEY_SPEC_LIST_EACH_BEGIN(spec_list, spec) do {    \
+    List<mrn_key_spec> spec_list_ = spec_list;                  \
+    List_iterator<mrn_key_spec> spec_iterator(spec_list_);      \
+    const mrn_key_spec *spec;                                   \
+    while ((spec = spec_iterator++))
+#  define MRN_KEY_SPEC_LIST_EACH_END()          \
+  } while (false)
+#endif
+
+#if MYSQL_VERSION_ID >= 80011 && !defined(MRN_MARIADB_P)
+#  define MRN_KEY_PART_SPEC_LIST_N_ELEMENTS(spec_list)      \
+  spec_list.size()
+#  define MRN_KEY_PART_SPEC_LIST_EACH_BEGIN(spec_list, spec) do {       \
+    for (size_t spec_i = 0; spec_i < spec_list.size(); ++spec_i) {      \
+      const Key_part_spec *spec = spec_list[spec_i];
+#  define MRN_KEY_PART_SPEC_LIST_EACH_END()     \
+    }                                           \
+  } while (false)
+#else
+#  define MRN_KEY_PART_SPEC_LIST_N_ELEMENTS(spec_list)      \
+  spec_list.elements
+#  define MRN_KEY_PART_SPEC_LIST_EACH_BEGIN(spec_list, spec) do {       \
+    List<Key_part_spec> spec_list_ = spec_list;                         \
+    List_iterator<Key_part_spec> spec_iterator(spec_list_);             \
+    const Key_part_spec *spec;                                          \
+    while ((spec = spec_iterator++))
+#  define MRN_KEY_PART_SPEC_LIST_EACH_END()     \
+  } while (false)
+#endif
+
+#if MYSQL_VERSION_ID >= 80011 && !defined(MRN_MARIADB_P)
+  typedef Foreign_key_spec mrn_foreign_key_spec;
+#else
+  typedef Foreign_key mrn_foreign_key_spec;
+#endif
+
+#if MYSQL_VERSION_ID >= 80011 && !defined(MRN_MARIADB_P)
+  class SELECT_LEX;
+  typedef SELECT_LEX mrn_select_lex;
+#else
+  typedef st_select_lex mrn_select_lex;
+#endif
+
+#if MYSQL_VERSION_ID >= 80011 && !defined(MRN_MARIADB_P)
+#  define MRN_FIELD_HAVE_AUTO_FLAGS
+#endif
+
+#if MYSQL_VERSION_ID >= 80011 && !defined(MRN_MARIADB_P)
+#  define MRN_OPEN_TABLE_DEF_USE_TABLE_DEFINITION
+#endif
+
+#if MYSQL_VERSION_ID >= 100007 && defined(MRN_MARIADB_P)
+#  define MRN_GET_ERR_MSG(code) my_get_err_msg(code)
+#else
+#  if MYSQL_VERSION_ID >= 80011 && !defined(MRN_MARIADB_P)
+#    define MRN_GET_ERR_MSG(code) ER_DEFAULT(code)
+#  else
+#    define MRN_GET_ERR_MSG(code) ER(code)
+#  endif
+#endif
+
+#define mrn_thd_set_ha_data(thd, hton, ha_data) \
+  *thd_ha_data(thd, hton) = ha_data
+// TODO
+/* #define mrn_thd_set_ha_data(thd, hton, ha_data) \ */
+/*    thd_set_ha_data(thd, hton, ha_data) */
+
+#if MYSQL_VERSION_ID >= 80011 && !defined(MRN_MARIADB_P)
+#  define mrn_destroy(pointer) destroy(pointer)
+#else
+#  define mrn_destroy(pointer) delete pointer
+#endif
